# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict, Union

import jax.numpy as jnp
from jax import Array, jit
from jax.typing import ArrayLike

from pyhgf.typing import NodeStructure


@partial(jit, static_argnames=("node_structure", "node_idx"))
def continuous_node_update(
    parameters_structure: Dict,
    time_step: float,
    node_idx: int,
    node_structure: NodeStructure,
    **args
) -> Dict:
    """Update the value and volatility parents of a continuous node.

    If the parents have value and/or volatility parents, they will be updated
    recursively.

    Updating the node's parents is a two step process:
    1. Update value parent(s) and their parents (if provided).
    2. Update volatility parent(s) and their parents (if provided).

    Then returns the new node structure.

    Parameters
    ----------
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    time_step :
        Interval between the previous time point and the current time point.
    node_idx :
        Pointer to the node that need to be updated. After continuous update, the
        parameters of value and volatility parents (if any) will be different.
    node_structure :
        Tuple of :py:class:`pyhgf.typing.Indexes` with same length than number of node.
        For each node, the index list value and volatility parents.

    Returns
    -------
    parameters_structure :
        The updated node structure.

    See Also
    --------
    update_continuous_input_parents, update_binary_input_parents

    """
    # using the current node index, unwrap parameters and parents
    node_parameters = parameters_structure[node_idx]
    value_parents_idx = node_structure[node_idx].value_parents
    volatility_parents_idx = node_structure[node_idx].volatility_parents

    # return here if no parents node are provided
    if (value_parents_idx is None) and (volatility_parents_idx is None):
        return parameters_structure

    pihat = node_parameters["pihat"]

    #######################################
    # Update the continuous value parents #
    #######################################
    if value_parents_idx is not None:
        vape = node_parameters["mu"] - node_parameters["muhat"]
        psis = node_parameters["psis"]

        for va_pa_idx, psi in zip(value_parents_idx, psis):
            # unpack the current parent's parameters with value and volatility parents
            va_pa_node_parameters = parameters_structure[va_pa_idx]
            va_pa_value_parents_idx = node_structure[va_pa_idx].value_parents
            va_pa_volatility_parents_idx = node_structure[va_pa_idx].volatility_parents

            # Compute new value for nu and pihat
            logvol = va_pa_node_parameters["omega"]

            # Look at the (optional) va_pa's volatility parents
            # and update logvol accordingly
            if va_pa_volatility_parents_idx is not None:
                for va_pa_vo_pa, k in zip(
                    va_pa_volatility_parents_idx, va_pa_node_parameters["kappas"]
                ):
                    logvol += k * parameters_structure[va_pa_vo_pa]["mu"]

            # Estimate new_nu
            nu = time_step * jnp.exp(logvol)
            new_nu = jnp.where(nu > 1e-128, nu, jnp.nan)

            pihat_pa, nu_pa = [1 / (1 / va_pa_node_parameters["pi"] + new_nu), new_nu]
            pi_pa = pihat_pa + psi**2 * pihat

            # Compute new muhat
            driftrate = va_pa_node_parameters["rho"]

            # Look at the (optional) va_pa's value parents
            # and update driftrate accordingly
            if va_pa_value_parents_idx is not None:
                for va_pa_va_pa in va_pa_value_parents_idx:
                    driftrate += psi * va_pa_va_pa["mu"]

            muhat_pa = va_pa_node_parameters["mu"] + time_step * driftrate
            mu_pa = muhat_pa + psi * pihat / pi_pa * vape

            # Update this parent's parameters
            parameters_structure[va_pa_idx]["pihat"] = pihat_pa
            parameters_structure[va_pa_idx]["pi"] = pi_pa
            parameters_structure[va_pa_idx]["muhat"] = muhat_pa
            parameters_structure[va_pa_idx]["mu"] = mu_pa
            parameters_structure[va_pa_idx]["nu"] = nu_pa

    #############################
    # Update volatility parents #
    #############################
    if volatility_parents_idx is not None:
        nu = node_parameters["nu"]
        kappas = node_parameters["kappas"]
        vope = (
            1 / node_parameters["pi"]
            + (node_parameters["mu"] - node_parameters["muhat"]) ** 2
        ) * node_parameters["pihat"] - 1

        for vo_pa_idx, kappa in zip(volatility_parents_idx, kappas):
            # unpack the current parent's parameters with value and volatility parents
            vo_pa_node_parameters = parameters_structure[vo_pa_idx]
            vo_pa_value_parents_idx = node_structure[vo_pa_idx].value_parents
            vo_pa_volatility_parents_idx = node_structure[vo_pa_idx].volatility_parents

            # Compute new value for nu and pihat
            logvol = vo_pa_node_parameters["omega"]

            # Look at the (optional) vo_pa's volatility parents
            # and update logvol accordingly
            if vo_pa_volatility_parents_idx is not None:
                for i, vo_pa_vo_pa in enumerate(vo_pa_volatility_parents_idx):
                    k = vo_pa_node_parameters["kappas"][i]
                    logvol += k * parameters_structure[vo_pa_vo_pa]["mu"]

            # Estimate new_nu
            new_nu = time_step * jnp.exp(logvol)
            new_nu = jnp.where(new_nu > 1e-128, new_nu, jnp.nan)

            pihat_pa, nu_pa = [1 / (1 / vo_pa_node_parameters["pi"] + new_nu), new_nu]

            pi_pa = pihat_pa + 0.5 * (kappa * nu * pihat) ** 2 * (
                1 + (1 - 1 / (nu * node_parameters["pi"])) * vope
            )
            pi_pa = jnp.where(pi_pa <= 0, jnp.nan, pi_pa)

            # Compute new muhat
            driftrate = vo_pa_node_parameters["rho"]

            # Look at the (optional) va_pa's value parents
            # and update driftrate accordingly
            if vo_pa_value_parents_idx is not None:
                for vo_pa_va_pa in vo_pa_value_parents_idx:
                    driftrate += psi * parameters_structure[vo_pa_va_pa]["mu"]

            muhat_pa = vo_pa_node_parameters["mu"] + time_step * driftrate
            mu_pa = muhat_pa + 0.5 * kappa * nu * pihat / pi_pa * vope

            # Update this parent's parameters
            parameters_structure[vo_pa_idx]["pihat"] = pihat_pa
            parameters_structure[vo_pa_idx]["pi"] = pi_pa
            parameters_structure[vo_pa_idx]["muhat"] = muhat_pa
            parameters_structure[vo_pa_idx]["mu"] = mu_pa
            parameters_structure[vo_pa_idx]["nu"] = nu_pa

    return parameters_structure


@partial(jit, static_argnames=("node_structure", "node_idx"))
def continuous_input_update(
    parameters_structure: Dict,
    time_step: float,
    node_idx: int,
    node_structure: NodeStructure,
    value: float,
) -> Dict:
    """Update the input node structure.

    This function is the entry level of the model fitting. It update the partents of
    the input node and then call :py:func:`pyhgf.jax.continuous_node_update` recursively
    to update the rest of the node structure.

    Parameters
    ----------
    value :
        The new observed value.
    time_step :
        Interval between the previous time point and the current time point.
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    node_structure :
        Tuple of :py:class:`pyhgf.typing.Indexes` with same length than number of node.
        For each node, the index list value and volatility parents.
    node_idx :
        Pointer to the node that need to be updated. After continuous update, the
        parameters of value and volatility parents (if any) will be different.

    Returns
    -------
    parameters_structure :
        The updated parameters structure.

    See Also
    --------
    continuous_node_update, update_binary_input_parents

    """
    # using the current node index, unwrap parameters and parents
    input_node_parameters = parameters_structure[node_idx]
    value_parents_idx = node_structure[node_idx].value_parents
    volatility_parents_idx = node_structure[node_idx].volatility_parents

    lognoise = input_node_parameters["omega"]

    if volatility_parents_idx is not None:
        for i, vo_pa_idx in enumerate(volatility_parents_idx):
            lognoise += (
                input_node_parameters["kappas"][i]
                * parameters_structure[vo_pa_idx]["mu"]
            )

    pihat = 1 / jnp.exp(lognoise)

    ########################
    # Update value parents #
    ########################
    if value_parents_idx is not None:
        # unpack the current parent's parameters with value and volatility parents
        va_pa_node_parameters = parameters_structure[value_parents_idx[0]]
        va_pa_value_parents_idx = node_structure[value_parents_idx[0]].value_parents
        va_pa_volatility_parents_idx = node_structure[
            value_parents_idx[0]
        ].volatility_parents

        vape = va_pa_node_parameters["mu"] - va_pa_node_parameters["muhat"]

        # Compute new value for nu and pihat
        logvol = va_pa_node_parameters["omega"]

        # Look at the (optional) va_pa's volatility parents
        # and update logvol accordingly
        if va_pa_volatility_parents_idx is not None:
            for va_pa_vo_pa, k in zip(
                va_pa_volatility_parents_idx, va_pa_node_parameters["kappas"]
            ):
                logvol += k * parameters_structure[va_pa_vo_pa]["mu"]

        # Estimate new_nu
        nu = time_step * jnp.exp(logvol)
        new_nu = jnp.where(nu > 1e-128, nu, jnp.nan)

        pihat_va_pa, nu_va_pa = [
            1 / (1 / va_pa_node_parameters["pi"] + new_nu),
            new_nu,
        ]
        pi_va_pa = pihat_va_pa + pihat

        # Compute new muhat
        driftrate = va_pa_node_parameters["rho"]

        # Look at the (optional) va_pa's value parents
        # and update driftrate accordingly
        if va_pa_value_parents_idx is not None:
            for va_pa_va_pa in va_pa_value_parents_idx:
                driftrate += (
                    va_pa_node_parameters["psis"]
                    * parameters_structure[va_pa_va_pa]["mu"]
                )

        muhat_va_pa = va_pa_node_parameters["mu"] + time_step * driftrate
        vape = value - muhat_va_pa
        mu_va_pa = muhat_va_pa + pihat / pi_va_pa * vape

        # update input node's parameters
        parameters_structure[value_parents_idx[0]]["pihat"] = pihat_va_pa
        parameters_structure[value_parents_idx[0]]["pi"] = pi_va_pa
        parameters_structure[value_parents_idx[0]]["muhat"] = muhat_va_pa
        parameters_structure[value_parents_idx[0]]["mu"] = mu_va_pa
        parameters_structure[value_parents_idx[0]]["nu"] = nu_va_pa

    #############################
    # Update volatility parents #
    #############################
    if volatility_parents_idx is not None:
        # unpack the current parent's parameters with value and volatility parents
        vo_pa_node_parameters = parameters_structure[volatility_parents_idx[0]]
        vo_pa_value_parents_idx = node_structure[
            volatility_parents_idx[0]
        ].value_parents
        vo_pa_volatility_parents_idx = node_structure[
            volatility_parents_idx[0]
        ].volatility_parents

        vope = (1 / pi_va_pa + (value - mu_va_pa) ** 2) * pihat - 1

        # Compute new value for nu and pihat
        logvol = vo_pa_node_parameters["omega"]

        # Look at the (optional) vo_pa's volatility parents
        # and update logvol accordingly
        if vo_pa_volatility_parents_idx is not None:
            for vo_pa_vo_pa in vo_pa_volatility_parents_idx:
                kappa = vo_pa_vo_pa[0]["kappas"]
                logvol += kappa * parameters_structure[vo_pa_vo_pa]["mu"]

        # Estimate new_nu
        nu = time_step * jnp.exp(logvol)
        new_nu = jnp.where(nu > 1e-128, nu, jnp.nan)

        pihat_vo_pa, nu_vo_pa = [
            1 / (1 / vo_pa_node_parameters["pi"] + new_nu),
            new_nu,
        ]

        pi_vo_pa = pihat_vo_pa + 0.5 * jnp.square(input_node_parameters["kappas"]) * (
            1 + vope
        )
        pi_vo_pa = jnp.where(pi_vo_pa <= 0, jnp.nan, pi_vo_pa)

        # Compute new muhat
        driftrate = vo_pa_node_parameters["rho"]

        # Look at the (optional) va_pa's value parents
        # and update driftrate accordingly
        if vo_pa_value_parents_idx is not None:
            for vo_pa_va_pa, p in zip(
                vo_pa_value_parents_idx, vo_pa_node_parameters["psis"]
            ):
                driftrate += p * parameters_structure[vo_pa_va_pa]["mu"]

        muhat_vo_pa = vo_pa_node_parameters["mu"] + time_step * driftrate
        mu_vo_pa = (
            muhat_vo_pa
            + jnp.multiply(0.5, input_node_parameters["kappas"]) / pi_vo_pa * vope
        )

        # Update this parent's parameters
        parameters_structure[volatility_parents_idx[0]]["pihat"] = pihat_vo_pa
        parameters_structure[volatility_parents_idx[0]]["pi"] = pi_vo_pa
        parameters_structure[volatility_parents_idx[0]]["muhat"] = muhat_vo_pa
        parameters_structure[volatility_parents_idx[0]]["mu"] = mu_vo_pa
        parameters_structure[volatility_parents_idx[0]]["nu"] = nu_vo_pa

    # store value and timestep in the node's parameters
    parameters_structure[node_idx]["time_step"] = time_step
    parameters_structure[node_idx]["value"] = value

    return parameters_structure


def gaussian_surprise(
    x: Union[float, ArrayLike],
    muhat: Union[float, ArrayLike],
    pihat: Union[float, ArrayLike],
) -> Array:
    r"""Surprise at an outcome under a Gaussian prediction.

    The surprise ellicited by an observation :math:`x` under a Gaussian distribution
    with expected mean :math:`\hat{\mu}` and expected precision :math:`\hat{\pi}` is
    given by:

    .. math::

       \frac{1}{2} \log(2 \pi) - \log(\hat{\pi}) + \hat{\pi}\sqrt{x - \hat{\mu}}

    where :math:`\pi` is the mathematical constant.

    Parameters
    ----------
    x :
        The outcome.
    muhat :
        The expected mean of the Gaussian distribution.
    pihat :
        The expected precision of the Gaussian distribution.

    Returns
    -------
    surprise :
        The Gaussian surprise.

    Examples
    --------
    >>> from pyhgf.continuous import gaussian_surprise
    >>> gaussian_surprise(x=2.0, muhat=0.0, pihat=1.0)
    `Array(2.9189386, dtype=float32, weak_type=True)`

    """
    return jnp.array(0.5) * (
        jnp.log(jnp.array(2.0) * jnp.pi)
        - jnp.log(pihat)
        + pihat * jnp.square(jnp.subtract(x, muhat))
    )



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict, Optional, Tuple

from jax import jit


def loop_inputs(
    parameters_structure: Dict,
    data: Tuple,
    update_sequence: Tuple,
    node_structure: Tuple,
) -> Tuple[Dict, Dict]:
    """Update the node structure after observing one new data point.

    One time step updating node structure and returning the new node structure with
    time, value and surprise stored in the input node.

    Parameters
    ----------
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    data :
        The current array element. Scan will iterate over a n x 2 Array with time steps
        and values (i.e. the input time series).
    update_sequence :
        The sequence of updates that will be applied to the node structure.
    node_structure :
        The node structure with the update sequence.

    """
    # Extract the current iteration variables (value and time)
    value, time_step = data

    # Fit the model with the current time and value variables, given model structure
    new_parameters_structure = apply_sequence(
        value=value,
        time_step=time_step,
        parameters_structure=parameters_structure,
        node_structure=node_structure,
        update_sequence=update_sequence,
    )

    return (
        new_parameters_structure,
        new_parameters_structure,
    )  # ("carryover", "accumulated")


@partial(jit, static_argnums=(3, 4))
def apply_sequence(
    value: Optional[float],
    time_step: Optional[float],
    parameters_structure: Dict,
    node_structure: Dict,
    update_sequence: Tuple,
) -> Dict:
    """Apply a predefinded update sequence to a node structure.

    Parameters
    ----------
    value :
        The new observation.
    time_step :
        The time interval between the new observation and the previous one.
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    node_structure :
        The node structure with the update sequence.
    update_sequence :
        The sequence of updates that will be applied to the node structure.

    """
    for sequence in update_sequence:
        node_idx, update_fn = sequence
        parameters_structure = update_fn(
            parameters_structure=parameters_structure,
            time_step=time_step,
            node_idx=node_idx,
            node_structure=node_structure,
            value=value,
        )
    return parameters_structure



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import pkg_resources  # type: ignore
from numpy import loadtxt

__version__ = "0.0.3"


def load_data(dataset: str):
    """Load dataset for continuous or binary HGF.

    Parameters
    ----------
    dataset : str
        The type of data to load. Can be `"continous"` or `"binary"`.

    Returns
    -------
    data : np.ndarray
        The data (a 1d timeseries).

    Notes
    -----
    The continuous time series is the standard USD-CHF conversion rates over time used
    in the Matlab examples. The binary dataset is from Iglesias et al. (2013) [#].

    References
    ----------
    .. [#] Iglesias, S., Mathys, C., Brodersen, K. H., Kasper, L., Piccirelli, M., den
    Ouden, H. E. M., & Stephan, K. E. (2013). Hierarchical Prediction Errors in Midbrain
    and Basal Forebrain during Sensory Learning. In Neuron (Vol. 80, Issue 2, pp.
    519–530). Elsevier BV. https://doi.org/10.1016/j.neuron.2013.09.009

    """
    if dataset == "continuous":
        data = loadtxt(pkg_resources.resource_filename("pyhgf", "/data/usdchf.dat"))
    elif dataset == "binary":
        data = loadtxt(
            pkg_resources.resource_filename("pyhgf", "/data/binary_input.dat")
        )
    else:
        raise ValueError("Invalid dataset argument. Should be 'continous' or 'binary'.")

    return data



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import unittest
from unittest import TestCase
import jax.numpy as jnp
from pyhgf.continuous import continuous_input_update, continuous_node_update
from pyhgf.structure import loop_inputs, apply_sequence
from pyhgf.typing import Indexes

class TestStructure(TestCase):

    def test_loop_inputs(self):
        """Test the loop_inputs function"""

        ###############################################
        # one value parent with one volatility parent #
        ###############################################
        input_node_parameters = {
            "omega": 1.0,
            "kappas": None,
            "psis": None,
            "surprise": 0.0,
            "time_step": 0.0,
            "value": 0.0,
        }
        node_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": (1.0,),
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 1.0,
        }

        node_structure = (
            Indexes((1,), None),
            Indexes(None, (2,)),
            Indexes(None, None),
        )
        parameters_structure = (
            input_node_parameters,
            node_parameters,
            node_parameters,
        )
        
        # create update sequence
        sequence1 = 0, continuous_input_update
        sequence2 = 1, continuous_node_update
        update_sequence = (sequence1, sequence2)

        # apply sequence
        new_parameters_structure = apply_sequence(
            node_structure=node_structure,
            parameters_structure=parameters_structure,
            update_sequence=update_sequence, 
            time_step=1.0,
            value=.2
            )

if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import itertools
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes

if TYPE_CHECKING:
    from graphviz.sources import Source

    from pyhgf.model import HGF


def plot_trajectories(
    hgf: "HGF",
    ci: bool = True,
    surprise: bool = True,
    figsize: Tuple[int, int] = (18, 9),
    axs: Optional[Union[List, Axes]] = None,
) -> Axes:
    r"""Plot the trajectories of the nodes' sufficient statistics and surprise.

    This function will plot :math:`\hat{mu}`, :math:`\¨hat{pi}` (converted into standard
    deviation) and the surprise at each level of the node structure.

    Parameters
    ----------
    hgf :
        Instance of the HGF model.
    ci :
        Show the uncertainty aroud the values estimates (standard deviation).
    surprise :
        If `True` plot each node's surprise together witt the sufficient statistics.
        If `False`, only the input node's surprise is depicted.
    figsize :
        The width and height of the figure. Defaults to `(18, 9)` for a 2-levels model,
        or to `(18, 12)` for a 3-levels model.
    axs :
        A list of Matplotlib axes instance where to draw the trajectories. This should
        correspond to the number of nodes in the structure. Default is `None` (create a
        new figure).

    Returns
    -------
    axs :
        The Matplotlib axes instances where to plot the trajectories.

    Examples
    --------
    Visualization of nodes' trajectories from a three-level continuous HGF model.

    .. plot::

        from pyhgf import load_data
        from pyhgf.model import HGF

        # Set up standard 3-level HGF for continuous inputs
        hgf = HGF(
            n_levels=3,
            model_type="continuous",
            initial_mu={"1": 1.04, "2": 1.0, "3": 1.0},
            initial_pi={"1": 1e4, "2": 1e1, "3": 1e1},
            omega={"1": -13.0, "2": -2.0, "3": -2.0},
            rho={"1": 0.0, "2": 0.0, "3": 0.0},
            kappas={"1": 1.0, "2": 1.0},
        )

        # Read USD-CHF data
        timeserie = load_data("continuous")

        # Feed input
        hgf.input_data(input_data=timeserie)

        # Plot
        hgf.plot_trajectories()

    Visualization of nodes' trajectories from a three-level binary HGF model.

    .. plot::

        from pyhgf import load_data
        from pyhgf.model import HGF
        import jax.numpy as jnp

        # Read binary input
        timeserie = load_data("binary")

        three_levels_hgf = HGF(
            n_levels=3,
            model_type="binary",
            initial_mu={"1": .0, "2": .5, "3": 0.},
            initial_pi={"1": .0, "2": 1e4, "3": 1e1},
            omega={"1": None, "2": -6.0, "3": -2.0},
            rho={"1": None, "2": 0.0, "3": 0.0},
            kappas={"1": None, "2": 1.0},
            eta0=0.0,
            eta1=1.0,
            pihat = jnp.inf,
        )

        # Feed input
        three_levels_hgf = three_levels_hgf.input_data(timeserie)

        # Plot
        three_levels_hgf.plot_trajectories()

    """
    trajectories_df = hgf.to_pandas()
    n_nodes = trajectories_df.columns.str.contains("_muhat").sum()
    palette = itertools.cycle(sns.color_palette())

    if axs is None:
        _, axs = plt.subplots(nrows=n_nodes + 1, figsize=figsize, sharex=True)

    # input node
    # ----------
    if hgf.model_type == "continuous":
        axs[n_nodes - 1].scatter(
            trajectories_df.time,
            trajectories_df.observation,
            s=3,
            label="Input",
            color="#2a2a2a",
            zorder=10,
            alpha=0.5,
        )
    elif hgf.model_type == "binary":
        axs[n_nodes - 1].scatter(
            x=trajectories_df.time,
            y=trajectories_df.observation,
            label="Input",
            color="#4c72b0",
            alpha=0.4,
            edgecolors="k",
            zorder=10,
        )

    # loop over the node idexes
    # -------------------------
    for i in range(1, n_nodes + 1):
        # use different colors for each nodes
        color = next(palette)

        # which ax instance to use
        ax_i = n_nodes - i

        # extract the sufficient statistics from the data frame
        mu = trajectories_df[f"x_{i}_muhat"]
        pi = trajectories_df[f"x_{i}_pihat"]

        # plotting mean
        axs[ax_i].plot(
            trajectories_df.time,
            mu,
            label=r"$\hat{\mu}$",
            color=color,
            linewidth=0.5,
            zorder=2,
        )

        # plotting standard deviation
        if ci is True:
            # if this is the first level of a binary model do not show CI
            if not (hgf.model_type == "binary") & (i == 1):
                sd = np.sqrt(1 / pi)
                axs[ax_i].fill_between(
                    x=trajectories_df.time,
                    y1=trajectories_df[f"x_{i}_muhat"] - sd,
                    y2=trajectories_df[f"x_{i}_muhat"] + sd,
                    alpha=0.4,
                    color=color,
                    zorder=2,
                )

        # plotting surprise
        if surprise:
            surprise_ax = axs[ax_i].twinx()
            surprise_ax.plot(
                trajectories_df.time,
                trajectories_df[f"x_{i}_surprise"],
                color="#2a2a2a",
                linewidth=0.5,
                linestyle="--",
                zorder=-1,
                label="Surprise",
            )
            surprise_ax.fill_between(
                x=trajectories_df.time,
                y1=trajectories_df[f"x_{i}_surprise"],
                y2=trajectories_df[f"x_{i}_surprise"].min(),
                color="#7f7f7f",
                alpha=0.2,
                zorder=-1,
            )
            surprise_ax.set_title(
                f"Node {i} surprise: {trajectories_df[f'x_{i}_surprise'].sum()}",
                loc="left",
            )
            surprise_ax.set_ylabel("Surprise")
        axs[ax_i].legend()
        axs[ax_i].set_ylabel(rf"$\mu_{i}$")

    # global surprise
    # ---------------
    surprise_ax = axs[n_nodes].twinx()
    surprise_ax.fill_between(
        x=trajectories_df.time,
        y1=trajectories_df.surprise,
        y2=trajectories_df.surprise.min(),
        label="Surprise",
        color="#7f7f7f",
        alpha=0.2,
    )
    surprise_ax.plot(
        trajectories_df.time,
        trajectories_df.surprise,
        color="#2a2a2a",
        linewidth=0.5,
        linestyle="--",
        zorder=-1,
        label="Surprise",
    )
    surprise_ax.set_title(
        f"Total surprise: {trajectories_df.surprise.sum()}", loc="left"
    )
    surprise_ax.set_ylabel("Surprise")
    surprise_ax.set_xlabel("Time")

    return axs


def plot_correlations(hgf: "HGF") -> Axes:
    """Plot the heatmap correlation of the sufficient statistics trajectories.

    Parameters
    ----------
    hgf :
        Instance of the HGF model.

    Returns
    -------
    axs :
        The Matplotlib ax instance containing the heatmap of parameters trajectories
        correlation.

    """
    trajectories_df = hgf.to_pandas()
    trajectories_df = pd.concat(
        [
            trajectories_df[["time", "observation", "surprise"]],
            trajectories_df.filter(regex="hat"),
        ],
        axis=1,
    )

    # rename columns with LateX expressions
    trajectories_df.columns = [
        r"$\hat{\mu}_" + f"{c[5]}$" if "muhat" in c else c
        for c in trajectories_df.columns
    ]
    trajectories_df.columns = [
        r"$\hat{\pi}_" + f"{c[5]}$" if "pihat" in c else c
        for c in trajectories_df.columns
    ]

    correlation_mat = trajectories_df.corr()
    ax = sns.heatmap(
        correlation_mat,
        annot=True,
        cmap="RdBu",
        vmin=-1,
        vmax=1,
        linewidths=2,
        square=True,
    )
    ax.set_title("Correlations between the model trajectories")

    return ax


def plot_network(hgf: "HGF") -> "Source":
    """Visualization of node network using GraphViz.

    Parameters
    ----------
    hgf :
        Instance of the HGF model containing a node structure.

    Notes
    -----
    This function requires [Graphviz](https://github.com/xflr6/graphviz) to be
    installed to work correctly.

    """
    try:
        import graphviz
    except ImportError:
        print(
            (
                "Graphviz is required to plot the nodes structure. "
                "See https://pypi.org/project/graphviz/"
            )
        )

    graphviz_structure = graphviz.Digraph("hgf-nodes", comment="Nodes structure")

    graphviz_structure.attr("node", shape="circle")

    # set input nodes
    list_of_input_idx = []
    for input_node in hgf.input_nodes_idx:
        list_of_input_idx.append(input_node.idx)
        graphviz_structure.node(
            f"x_{input_node.idx}",
            label=f"{input_node.kind.capitalize()[0]}I - {input_node.idx}",
            style="filled",
            shape="octagon",
        )

    # create the rest of nodes
    for i in range(len(hgf.node_structure)):
        # only if node is not an input node
        if i not in list_of_input_idx:
            graphviz_structure.node(f"x_{i}", label=str(i), shape="circle")

    # connect value parents
    for i, idx in enumerate(hgf.node_structure):
        value_parents, _ = idx

        if value_parents is not None:
            for value_parents_idx in value_parents:
                graphviz_structure.edge(
                    f"x_{value_parents_idx}",
                    f"x_{i}",
                )

    # connect volatility parents
    for i, idx in enumerate(hgf.node_structure):
        _, volatility_parents = idx

        if volatility_parents is not None:
            for volatility_parents_idx in volatility_parents:
                graphviz_structure.edge(
                    f"x_{volatility_parents_idx}",
                    f"x_{i}",
                    color="gray",
                    style="dashed",
                    arrowhead="dot",
                )

    # unflat the structure to better handle large/uneven networks
    graphviz_structure = graphviz_structure.unflatten(stagger=3)

    return graphviz_structure



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import unittest
from unittest import TestCase
from jax.tree_util import Partial

import jax.numpy as jnp
from jax.lax import scan

from pyhgf import load_data
from pyhgf.binary import (
    binary_input_update,
    binary_node_update,
    binary_surprise,
    gaussian_density,
    sgm
)
from pyhgf.continuous import continuous_node_update
from pyhgf.structure import apply_sequence, loop_inputs
from pyhgf.typing import Indexes



class Testbinary(TestCase):

    def test_gaussian_density(self):
        surprise = gaussian_density(
            x=jnp.array([1.0, 1.0]),
            mu=jnp.array([0.0, 0.0]),
            pi=jnp.array([1.0, 1.0]),
        )
        assert jnp.all(jnp.isclose(surprise, 0.24197073))

    def test_sgm(self):
        assert jnp.all(jnp.isclose(sgm(jnp.array([.3, .3])), 0.5744425))

    def test_binary_surprise(self):
        surprise = binary_surprise(
            x=jnp.array([1.0]),
            muhat=jnp.array([0.2]),
        )
        assert jnp.all(jnp.isclose(surprise, 1.609438))
        
    def test_update_binary_input_parents(self):

        ##########################
        # three level binary HGF #
        ##########################
        input_node_parameters = {
            "pihat": jnp.inf,
            "eta0": 0.0,
            "eta1": 1.0,
            "surprise": 0.0,
            "time_step": 0.0,
            "value": 0.0,
        }
        node_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": (1.0,),
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 0.0,
        }

        node_structure = (
            Indexes((1,), None),
            Indexes((2,), None),
            Indexes(None, (3,)),
            Indexes(None, None),
        )
        parameters_structure = (
            input_node_parameters,
            node_parameters,
            node_parameters,
            node_parameters,
        )
        
        # create update sequence
        sequence1 = 0, binary_input_update
        sequence2 = 1, binary_node_update
        sequence3 = 2, continuous_node_update
        update_sequence = (sequence1, sequence2, sequence3)

        # apply sequence
        new_parameters_structure = apply_sequence(
            node_structure=node_structure,
            parameters_structure=parameters_structure,
            update_sequence=update_sequence, 
            time_step=1.0,
            value=1.0
            )
        assert new_parameters_structure[0]["surprise"] == 0.31326166

        # use scan
        timeserie = load_data("binary")

        # Create the data (value and time steps vectors)
        data = jnp.array([timeserie, jnp.ones(len(timeserie), dtype=int)]).T


        # create the function that will be scaned
        scan_fn = Partial(
            loop_inputs, 
            update_sequence=update_sequence, 
            node_structure=node_structure
            )

        # Run the entire for loop
        last, final = scan(scan_fn, parameters_structure, data)

if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
import pytensor.tensor as pt
from jax import grad, jit
from jax.tree_util import Partial
from jax.typing import ArrayLike
from pytensor.graph import Apply, Op

from pyhgf.model import HGF


def hgf_logp(
    omega_1: Union[np.ndarray, ArrayLike, float],
    omega_2: Union[np.ndarray, ArrayLike, float],
    omega_3: Union[np.ndarray, ArrayLike, float],
    omega_input: Union[np.ndarray, ArrayLike, float],
    rho_1: Union[np.ndarray, ArrayLike, float],
    rho_2: Union[np.ndarray, ArrayLike, float],
    rho_3: Union[np.ndarray, ArrayLike, float],
    pi_1: Union[np.ndarray, ArrayLike, float],
    pi_2: Union[np.ndarray, ArrayLike, float],
    pi_3: Union[np.ndarray, ArrayLike, float],
    mu_1: Union[np.ndarray, ArrayLike, float],
    mu_2: Union[np.ndarray, ArrayLike, float],
    mu_3: Union[np.ndarray, ArrayLike, float],
    kappa_1: Union[np.ndarray, ArrayLike, float],
    kappa_2: Union[np.ndarray, ArrayLike, float],
    input_data: List[np.ndarray],
    response_function: Callable,
    model_type: str,
    n_levels: int,
    response_function_parameters: Tuple = (),
    time: Optional[List] = None,
) -> float:
    r"""Log probability from HGF model(s) given input data and parameter(s).

    This function support broadcasting along the first axis:
    - If the input data contains many time series, the function will automatically
    create the corresponding number of HGF models and fit them separately.
    - If a single input data is provided but some parameters have array-like inputs, the
    number of HGF models will match the length of the arrays, using the value *i*
    for the *i* th model. When floats are provided for some parameters, the same value
    will be used for all HGF models.
    - If multiple input data are provided with array-like inputs for some parameter, the
    function will create and fit the models separately using the value *i* for the
    *i* th model.


    Parameters
    ----------
    omega_1 :
        The :math:`\omega` parameter, or *evolution rate*, at the first level of the
        HGF. This parameter represents the tonic part of the variance (the part that is
        not inherited from parents nodes).
    omega_2 :
        The :math:`\omega` parameter, or *evolution rate*, at the second level of the
        HGF. This parameter represents the tonic part of the variance (the part that is
        not inherited from parents nodes).
    omega_3 :
        The :math:`\omega` parameter, or *evolution rate*, at the third level of the
        HGF. This parameter represents the tonic part of the variance (the part that is
        not inherited from parents nodes). The value of this parameter will be ignored
        when using a 2-levels HGF (`n_levels=2`).
    omega_input :
        Represent the noise associated with the input observation.
    rho_1 :
        The :math:`\rho` parameter at the first level of the HGF. This parameter
        represents the drift of the random walk.
    rho_2 :
        The :math:`\rho` parameter at the second level of the HGF. This parameter
        represents the drift of the random walk.
    rho_3 :
        The :math:`\rho` parameter at the first level of the HGF. This parameter
        represents the drift of the random walk. The value of this parameter will be
        ignored when using a 2-levels HGF (`n_levels=2`).
    pi_1 :
        The :math:`\pi` parameter, or *precision*, at the first level of the HGF.
    pi_2 :
        The :math:`\pi` parameter, or *precision*, at the second level of the HGF.
    pi_3 :
        The :math:`\pi` parameter, or *precision*, at the third level of the HGF. The
        value of this parameter will be ignored when using a 2-levels HGF
        (`n_levels=2`).
    mu_1 :
        The :math:`\mu` parameter, or *mean*, at the first level of the HGF.
    mu_2 :
        The :math:`\mu` parameter, or *mean*, at the second level of the HGF.
    mu_3 :
        The :math:`\mu` parameter, or *mean*, at the third level of the HGF. The value
        of this parameter will be ignored when using a 2-levels HGF (`n_levels=2`).
    kappa_1 :
        The value of the :math:`\\kappa` parameter at the first level of the HGF. Kappa
        represents the phasic part of the variance (the part that is affected by the
        parents nodes) and will defines the strenght of the connection between the node
        and the parent node. Often fixed to `1`.
    kappa_2 :
        The value of the :math:`\\kappa` parameter at the second level of the HGF. Kappa
        represents the phasic part of the variance (the part that is affected by the
        parents nodes) and will defines the strenght of the connection between the node
        and the parent node. Often fixed to `1`. The value of this parameter will be
        ignored when using a 2-levels HGF (`n_levels=2`).
    input_data :
        List of input data. When `n` models should be fitted, the list contains `n` 1d
        Numpy arrays. By default, the associated time vector is the integers vector
        starting at `0`. A different time vector can be passed to the `time` argument.
    response_function :
        The response function to use to compute the model surprise.
    model_type :
        The model type to use (can be "continuous" or "binary").
    n_levels :
        The number of hierarchies in the perceptual model (can be `2` or `3`). If
        `None`, the nodes hierarchy is not created and might be provided afterward
        using :py:meth:`pyhgf.model.HGF.add_nodes`.
    response_function_parameters :
        Additional parameters to the response function.
    time :
        List of 1d Numpy arrays containing the time vectors for each input time series.
        If one of the list item is `None`, or if `None` is provided instead, the time
        vector will defaults to integers vector starting at 0.

    Returns
    -------
    log_prob :
        The sum of the log-probabilities (or negative surprise).

    """
    # number of models
    n = len(input_data)

    # Broadcast inputs to an array with length n>=1
    (
        _omega_1,
        _omega_2,
        _omega_3,
        _omega_input,
        _rho_1,
        _rho_2,
        _rho_3,
        _pi_1,
        _pi_2,
        _pi_3,
        _mu_1,
        _mu_2,
        _mu_3,
        _kappa_1,
        _kappa_2,
        _,
    ) = jnp.broadcast_arrays(
        omega_1,
        omega_2,
        omega_3,
        omega_input,
        rho_1,
        rho_2,
        rho_3,
        pi_1,
        pi_2,
        pi_3,
        mu_1,
        mu_2,
        mu_3,
        kappa_1,
        kappa_2,
        jnp.zeros(n),
    )

    # if no time vectors provided, set it to None (will defaults to integers vectors)
    # otherwise check consistency with the input data
    if time is None:
        time = [None] * n
    else:
        assert len(time) == n

    surprise = 0.0

    # Fitting n HGF models to the n datasets
    for i in range(n):
        # Format HGF parameters
        initial_mu: Dict = {
            "1": _mu_1[i],
            "2": _mu_2[i],
            "3": _mu_3[i],
        }
        initial_pi: Dict = {
            "1": _pi_1[i],
            "2": _pi_2[i],
            "3": _pi_3[i],
        }
        omega: Dict = {
            "1": _omega_1[i],
            "2": _omega_2[i],
            "3": _omega_3[i],
        }
        rho: Dict = {"1": _rho_1[i], "2": _rho_2[i], "3": _rho_3[i]}
        kappas: Dict = {"1": _kappa_1[i], "2": _kappa_2[i]}

        surprise = surprise + (
            HGF(
                initial_mu=initial_mu,
                initial_pi=initial_pi,
                omega=omega,
                omega_input=_omega_input[i],
                rho=rho,
                kappas=kappas,
                model_type=model_type,
                n_levels=n_levels,
                eta0=0.0,
                eta1=1.0,
                pihat=jnp.inf,
                verbose=False,
            )
            .input_data(input_data=input_data[i], time=time[i])
            .surprise(
                response_function=response_function,
                response_function_parameters=response_function_parameters,
            )
        )

    # Return the sum of the log probabilities (negative surprise)
    return -surprise


class HGFLogpGradOp(Op):
    """Gradient Op for the HGF distribution."""

    def __init__(
        self,
        input_data: List = [],
        time: Optional[List] = None,
        model_type: str = "continous",
        n_levels: int = 2,
        response_function: Optional[Callable] = None,
        response_function_parameters: Tuple = (),
    ):
        """Initialize function.

        Parameters
        ----------
        input_data :
            List of input data. When `n` models should be fitted, the list contains `n`
            1d Numpy arrays. By default, the associated time vector is the integers
            vector starting at `0`. A different time vector can be passed to the `time`
            argument.
        time :
            List of 1d Numpy arrays containing the time vectors for each input time
            series. If one of the list item is `None`, or if `None` is provided instead,
            the time vector will defaults to integers vector starting at 0.
        model_type :
            The model type to use (can be "continuous" or "binary").
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterward
            using `add_nodes()`.
        response_function :
            The response function to use to compute the model surprise.
        response_function_parameters :
            Additional parameters to the response function.

        """
        self.grad_logp = jit(
            grad(
                Partial(
                    hgf_logp,
                    n_levels=n_levels,
                    input_data=input_data,
                    time=time,
                    response_function=response_function,
                    model_type=model_type,
                    response_function_parameters=response_function_parameters,
                ),
                argnums=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14],
            )
        )

    def make_node(
        self,
        omega_1=np.array(-3.0),
        omega_2=np.array(-3.0),
        omega_3=np.array(0.0),
        omega_input=np.log(1e-4),
        rho_1=np.array(0.0),
        rho_2=np.array(0.0),
        rho_3=np.array(0.0),
        pi_1=np.array(1e4),
        pi_2=np.array(1e1),
        pi_3=np.array(0.0),
        mu_1=np.array(0.0),
        mu_2=np.array(0.0),
        mu_3=np.array(0.0),
        kappa_1=np.array(1.0),
        kappa_2=np.array(0.0),
    ):
        """Initialize node structure."""
        # Convert our inputs to symbolic variables
        inputs = [
            pt.as_tensor_variable(omega_1),
            pt.as_tensor_variable(omega_2),
            pt.as_tensor_variable(omega_3),
            pt.as_tensor_variable(omega_input),
            pt.as_tensor_variable(rho_1),
            pt.as_tensor_variable(rho_2),
            pt.as_tensor_variable(rho_3),
            pt.as_tensor_variable(pi_1),
            pt.as_tensor_variable(pi_2),
            pt.as_tensor_variable(pi_3),
            pt.as_tensor_variable(mu_1),
            pt.as_tensor_variable(mu_2),
            pt.as_tensor_variable(mu_3),
            pt.as_tensor_variable(kappa_1),
            pt.as_tensor_variable(kappa_2),
        ]
        # This `Op` will return one gradient per input. For simplicity, we assume
        # each output is of the same type as the input. In practice, you should use
        # the exact dtype to avoid overhead when saving the results of the computation
        # in `perform`
        outputs = [inp.type() for inp in inputs]
        return Apply(self, inputs, outputs)

    def perform(self, node, inputs, outputs):
        """Perform node operations."""
        (
            grad_omega_1,
            grad_omega_2,
            grad_omega_3,
            grad_omega_input,
            grad_rho_1,
            grad_rho_2,
            grad_rho_3,
            grad_pi_1,
            grad_pi_2,
            grad_pi_3,
            grad_mu_1,
            grad_mu_2,
            grad_mu_3,
            grad_kappa_1,
            grad_kappa_2,
        ) = self.grad_logp(*inputs)

        outputs[0][0] = np.asarray(grad_omega_1, dtype=node.outputs[0].dtype)
        outputs[1][0] = np.asarray(grad_omega_2, dtype=node.outputs[1].dtype)
        outputs[2][0] = np.asarray(grad_omega_3, dtype=node.outputs[2].dtype)
        outputs[3][0] = np.asarray(grad_omega_input, dtype=node.outputs[3].dtype)
        outputs[4][0] = np.asarray(grad_rho_1, dtype=node.outputs[4].dtype)
        outputs[5][0] = np.asarray(grad_rho_2, dtype=node.outputs[5].dtype)
        outputs[6][0] = np.asarray(grad_rho_3, dtype=node.outputs[6].dtype)
        outputs[7][0] = np.asarray(grad_pi_1, dtype=node.outputs[7].dtype)
        outputs[8][0] = np.asarray(grad_pi_2, dtype=node.outputs[8].dtype)
        outputs[9][0] = np.asarray(grad_pi_3, dtype=node.outputs[9].dtype)
        outputs[10][0] = np.asarray(grad_mu_1, dtype=node.outputs[10].dtype)
        outputs[11][0] = np.asarray(grad_mu_2, dtype=node.outputs[11].dtype)
        outputs[12][0] = np.asarray(grad_mu_3, dtype=node.outputs[12].dtype)
        outputs[13][0] = np.asarray(grad_kappa_1, dtype=node.outputs[13].dtype)
        outputs[14][0] = np.asarray(grad_kappa_2, dtype=node.outputs[14].dtype)


class HGFDistribution(Op):
    """The HGF distribution PyMC >= 5.0 compatible.

    Examples
    --------
    This class should be used in the context of a PyMC probabilistic model to estimate
    one or many parameter-s probability density with MCMC sampling.

    >>> import arviz as az
    >>> import numpy as np
    >>> import pymc as pm
    >>> from math import log
    >>> from pyhgf import load_data
    >>> from pyhgf.distribution import HGFDistribution
    >>> from pyhgf.response import total_gaussian_surprise

    Create the data (value and time vectors)

    >>> timeserie = load_data("continuous")

    We create the PyMC distribution here, specifying the type of model and response
    function we want to use (i.e simple gaussian surprise).

    >>> hgf_logp_op = HGFDistribution(
    >>>     n_levels=2,
    >>>     input_data=timeserie,
    >>>     response_function=total_gaussian_surprise,
    >>> )

    Create the PyMC model
    Here we are fixing all the paramters to arbitrary values and sampling the
    posterior probability density of Omega in the second level, using a normal
    distribution as prior : ω_2 ~ N(-11, 2).

    >>> with pm.Model() as model:
    >>>     ω_2 = pm.Normal("ω_2", -11.0, 2)
    >>>     pm.Potential(
    >>>         "hhgf_loglike",
    >>>         hgf_logp_op(
    >>>             omega_1=0.0,
    >>>             omega_2=ω_2,
    >>>             omega_input=log(1e-4),
    >>>             rho_1=0.0,
    >>>             rho_2=0.0,
    >>>             pi_1=1e4,
    >>>             pi_2=1e1,
    >>>             mu_1=timeserie[0],
    >>>             mu_2=0.0,
    >>>             kappa_1=1.0,
    >>>         ),
    >>>     )

    Sample the model - Using 4 chain, 4 cores and 1000 warmups.

    >>> with model:
    >>>     hgf_samples = pm.sample(chains=4, cores=4, tune=1000)

    Print a summary of the results using Arviz

    >>> az.summary(hgf_samples, var_names="ω_2")
    ===  =======  =====  =======  =======  =====  =====  ====  ====  =
    ω_2  -12.846  1.305  -15.466  -10.675  0.038  0.027  1254  1726  1
    ===  =======  =====  =======  =======  =====  =====  ====  ====  =

    """

    def __init__(
        self,
        input_data: List[np.ndarray] = [],
        time: Optional[List] = None,
        model_type: str = "continuous",
        n_levels: int = 2,
        response_function: Optional[Callable] = None,
        response_function_parameters: Tuple = (),
    ):
        """Distribution initialization.

        Parameters
        ----------
        input_data :
            List of input data. When `n` models should be fitted, the list contains `n`
            1d Numpy arrays. By default, the associated time vector is the integers
            vector starting at `0`. A different time vector can be passed to the `time`
            argument.
        time :
            List of 1d Numpy arrays containing the time vectors for each input time
            series. If one of the list item is `None`, or if `None` is provided instead,
            the time vector will defaults to integers vector starting at 0.
        model_type :
            The model type to use (can be "continuous" or "binary").
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterward
            using `add_nodes()`.
        response_function :
            The response function to use to compute the model surprise.
        response_function_parameters :
            Additional parameters to the response function.

        """
        # The value function
        self.hgf_logp = jit(
            Partial(
                hgf_logp,
                n_levels=n_levels,
                input_data=input_data,
                time=time,
                response_function=response_function,
                model_type=model_type,
                response_function_parameters=response_function_parameters,
            ),
        )

        # The gradient function
        self.hgf_logp_grad_op = HGFLogpGradOp(
            input_data=input_data,
            time=time,
            model_type=model_type,
            n_levels=n_levels,
            response_function=response_function,
            response_function_parameters=response_function_parameters,
        )

    def make_node(
        self,
        omega_1,
        omega_2,
        omega_3,
        omega_input,
        rho_1,
        rho_2,
        rho_3,
        pi_1,
        pi_2,
        pi_3,
        mu_1,
        mu_2,
        mu_3,
        kappa_1,
        kappa_2,
    ):
        """Convert inputs to symbolic variables."""
        inputs = [
            pt.as_tensor_variable(omega_1),
            pt.as_tensor_variable(omega_2),
            pt.as_tensor_variable(omega_3),
            pt.as_tensor_variable(omega_input),
            pt.as_tensor_variable(rho_1),
            pt.as_tensor_variable(rho_2),
            pt.as_tensor_variable(rho_3),
            pt.as_tensor_variable(pi_1),
            pt.as_tensor_variable(pi_2),
            pt.as_tensor_variable(pi_3),
            pt.as_tensor_variable(mu_1),
            pt.as_tensor_variable(mu_2),
            pt.as_tensor_variable(mu_3),
            pt.as_tensor_variable(kappa_1),
            pt.as_tensor_variable(kappa_2),
        ]
        # Define the type of the output returned by the wrapped JAX function
        outputs = [pt.dscalar()]
        return Apply(self, inputs, outputs)

    def perform(self, node, inputs, outputs):
        """Run the function forward."""
        result = self.hgf_logp(*inputs)
        outputs[0][0] = np.asarray(result, dtype=node.outputs[0].dtype)

    def grad(self, inputs, output_gradients):
        """Gradient of the function."""
        (
            grad_omega_1,
            grad_omega_2,
            grad_omega_3,
            grad_omega_input,
            grad_rho_1,
            grad_rho_2,
            grad_rho_3,
            grad_pi_1,
            grad_pi_2,
            grad_pi_3,
            grad_mu_1,
            grad_mu_2,
            grad_mu_3,
            grad_kappa_1,
            grad_kappa_2,
        ) = self.hgf_logp_grad_op(*inputs)
        # If there are inputs for which the gradients will never be needed or cannot
        # be computed, `aesara.gradient.grad_not_implemented` should  be used as the
        # output gradient for that input.
        output_gradient = output_gradients[0]
        return [
            output_gradient * grad_omega_1,
            output_gradient * grad_omega_2,
            output_gradient * grad_omega_3,
            output_gradient * grad_omega_input,
            output_gradient * grad_rho_1,
            output_gradient * grad_rho_2,
            output_gradient * grad_rho_3,
            output_gradient * grad_pi_1,
            output_gradient * grad_pi_2,
            output_gradient * grad_pi_3,
            output_gradient * grad_mu_1,
            output_gradient * grad_mu_2,
            output_gradient * grad_mu_3,
            output_gradient * grad_kappa_1,
            output_gradient * grad_kappa_2,
        ]



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from math import log
from typing import Callable, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
import pandas as pd
from jax.lax import scan
from jax.tree_util import Partial
from jax.typing import ArrayLike

from pyhgf.binary import binary_input_update, binary_node_update
from pyhgf.continuous import (
    continuous_input_update,
    continuous_node_update,
    gaussian_surprise,
)
from pyhgf.plots import plot_correlations, plot_network, plot_trajectories
from pyhgf.response import first_level_binary_surprise, first_level_gaussian_surprise
from pyhgf.structure import loop_inputs
from pyhgf.typing import Indexes, InputIndexes, NodeStructure


class HGF(object):
    r"""The two-level and three-level Hierarchical Gaussian Filters (HGF).

    This class use pre-made nodes structures that corresponds to the most widely used
    HGF. The inputs can be continuous or binary.

    Attributes
    ----------
    input_nodes_idx :
        Indexes of the input nodes. Defaults to `(0,)` if the network only has one input
        node.
    model_type :
        The model implemented (can be `"continuous"`, `"binary"` or `"custom"`).
    n_levels :
        The number of hierarchies in the model, including the input vector. Cannot be
        less than 2.
    node_structure :
        A tuple of :py:class:`pyhgf.typing.Indexes` representing the nodes hierarchy.
    node_trajectories :
        The parameter structure that incluse the consecutive updates at each new
        observation.
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes.
        `"kappas"` is the volatility coupling strength. It should have same length
        than the volatility parents' indexes.
    verbose : bool
        Verbosity level.

    """

    def __init__(
        self,
        n_levels: Optional[int] = 2,
        model_type: str = "continuous",
        initial_mu: Dict = {
            "1": 0.0,
            "2": 0.0,
            "3": 0.0,
        },
        initial_pi: Dict = {
            "1": 1.0,
            "2": 1.0,
            "3": 1.0,
        },
        omega_input: Union[float, np.ndarray, ArrayLike] = log(1e-4),
        omega: Dict = {
            "1": -3.0,
            "2": -3.0,
            "3": -3.0,
        },
        kappas: Dict = {"1": 1.0, "2": 0.0},
        eta0: Union[float, np.ndarray, ArrayLike] = 0.0,
        eta1: Union[float, np.ndarray, ArrayLike] = 1.0,
        pihat: Union[float, np.ndarray, ArrayLike] = jnp.inf,
        rho: Dict = {
            "1": 0.0,
            "2": 0.0,
            "3": 0.0,
        },
        verbose: bool = True,
    ):
        r"""Parameterization of the HGF model.

        Parameters
        ----------
        eta0 :
            The first categorical value of the binary node. Defaults to `0.0`. Only
            relevant if `model_type="binary"`.
        eta1 :
            The second categorical value of the binary node. Defaults to `0.0`. Only
            relevant if `model_type="binary"`.
        initial_mu :
            Dictionary containing the initial values for the :math:`\\mu` parameter at
            different levels of the hierarchy. Defaults set to `0.0` at all levels.
        initial_pi :
            Dictionary containing the initial values for the :math:`\\pi` parameter at
            different levels of the hierarchy. Pis values encode the precision of the
            values at each level (Var = 1/pi) Defaults set to `1.0` at all levels.
        kappas :
            Dictionary containing the initial values for the :math:`\\kappa` parameter
            at different levels of the hierarchy. Kappa represents the phasic part of
            the variance (the part that is affected by the parents nodes) and will
            defines the strenght of the connection between the node and the parent
            node. Often fixed to `1`. Defaults set to `1` at all levels.
        model_type : str
            The model type to use (can be `"continuous"` or `"binary"`).
        n_levels :
            The number of hierarchies in the perceptual model (can be `2` or `3`). If
            `None`, the nodes hierarchy is not created and might be provided afterward
            using `add_nodes()`. Defaults to `2` for a 2-level HGF.
        omega :
            Dictionary containing the initial values for the :math:`\\omega` parameter
            at different levels of the hierarchy. :math:`\\omega` represents the tonic
            part of the variance (the part that is not affected by the parent node).
            Defaults set to `-3.0` at all levels. This parameter is only required for
            Gaussian Random Walk nodes.
        omega_input :
            Default value sets to `log(1e-4)`. Represents the noise associated with
            the input.
        pihat :
            The precision of the binary input node. Default to `jnp.inf`. Only relevant
            if `model_type="binary"`.
        rho :
            Dictionary containing the initial values for the :math:`\\rho` parameter at
            different levels of the hierarchy. Rho represents the drift of the random
            walk. This parameter is only required for Gaussian Random Walk nodes.
            Defaults set all entries to `0` according to the number of required levels.
        verbose :
            Verbosity of the methods for model creation and fitting. Defaults to `True`.

        """
        self.model_type = model_type
        self.verbose = verbose
        self.n_levels = n_levels
        self.node_structure: NodeStructure
        self.node_trajectories: Dict
        self.parameters_structure: Dict
        self.input_nodes_idx: Tuple[InputIndexes, ...]

        if model_type in ["continuous", "binary"]:
            if self.verbose:
                print(
                    (
                        f"Creating a {self.model_type} Hierarchical Gaussian Filter "
                        f"with {self.n_levels} levels."
                    )
                )

            #########
            # Input #
            #########
            if model_type == "continuous":
                self.add_input_node(kind="continuous", omega_input=omega_input)
            elif model_type == "binary":
                self.add_input_node(kind="binary", eta0=eta0, eta1=eta1, pihat=pihat)

            #########
            # x - 1 #
            #########
            self.add_value_parent(
                children_idxs=[0],
                value_coupling=1.0,
                mu=initial_mu["1"],
                pi=initial_pi["1"],
                omega=omega["1"] if self.model_type != "binary" else np.nan,
                rho=rho["1"] if self.model_type != "binary" else np.nan,
            )

            #########
            # x - 2 #
            #########
            if model_type == "continuous":
                self.add_volatility_parent(
                    children_idxs=[1],
                    volatility_coupling=kappas["1"],
                    mu=initial_mu["2"],
                    pi=initial_pi["2"],
                    omega=omega["2"],
                    rho=rho["2"],
                )
            elif model_type == "binary":
                self.add_value_parent(
                    children_idxs=[1],
                    value_coupling=1.0,
                    mu=initial_mu["2"],
                    pi=initial_pi["2"],
                    omega=omega["2"],
                    rho=rho["2"],
                )

            #########
            # x - 3 #
            #########
            if self.n_levels == 3:
                self.add_volatility_parent(
                    children_idxs=[2],
                    volatility_coupling=kappas["2"],
                    mu=initial_mu["3"],
                    pi=initial_pi["3"],
                    omega=omega["3"],
                    rho=rho["3"],
                )

            # get the update sequence from structure
            self.update_sequence = self.get_update_sequence()

    def input_data(
        self,
        input_data: np.ndarray,
        time: Optional[np.ndarray] = None,
    ):
        """Add new observations.

        Parameters
        ----------
        input_data :
            The new observations.
        time :
            Time vector (optional). If `None`, the time vector will defaults to
            `np.arange(0, len(input_data))`.

        """
        if self.verbose:
            print((f"Add {len(input_data)} new {self.model_type} observations."))
        if time is None:
            time = np.ones(len(input_data), dtype=int)  # time step vector

        # create the function that will be scaned
        scan_fn = Partial(
            loop_inputs,
            update_sequence=self.update_sequence,
            node_structure=self.node_structure,
        )

        # create the input array (data, time)
        data = jnp.array([input_data, time], dtype=float).T

        # Run the entire for loop
        # this is where the model loop over the input time series
        # at each input the node structure is traversed and beliefs are updated
        # using precision-weighted prediction errors
        _, node_trajectories = scan(scan_fn, self.parameters_structure, data)

        # the node structure at each value update
        self.node_trajectories = node_trajectories

        return self

    def plot_trajectories(self, **kwargs):
        """Plot the parameters trajectories."""
        return plot_trajectories(hgf=self, **kwargs)

    def plot_correlations(self):
        """Plot the heatmap of cross-trajectories correlation."""
        return plot_correlations(hgf=self)

    def plot_network(self):
        """Visualization of node network using GraphViz."""
        return plot_network(hgf=self)

    def surprise(
        self,
        response_function: Optional[Callable] = None,
        response_function_parameters: Tuple = (),
    ):
        """Surprise (negative log probability) under new observations.

        The surprise depends on the input data, the model parameters, the response
        function and its additional parameters(optional).

        Parameters
        ----------
        response_function :
            The response function to use to compute the model surprise. If `None`
            (default), return the sum of gaussian surprise if `model_type=="continuous"`
            or the sum of the binary surprise if `model_type=="binary"`.
        response_function_parameters :
            Additional parameters to the response function (optional) .

        Returns
        -------
        surprise :
            The model surprise given the input data and the response function.

        """
        if response_function is None:
            response_function = (
                first_level_gaussian_surprise
                if self.model_type == "continuous"
                else first_level_binary_surprise
            )

        return response_function(
            hgf=self,
            response_function_parameters=response_function_parameters,
        )

    def to_pandas(self) -> pd.DataFrame:
        """Export the nodes trajectories and surprise as a Pandas data frame.

        Returns
        -------
        structure_df :
            Pandas data frame with the time series of the sufficient statistics and
            surprise of each node in the structure.

        """
        # input parameters
        structure_df = pd.DataFrame(
            {
                "time_steps": self.node_trajectories[0]["time_step"],
                "time": jnp.cumsum(self.node_trajectories[0]["time_step"]),
                "observation": self.node_trajectories[0]["value"],
            }
        )

        # loop over nodes and store sufficient statistics with surprise
        for i in range(1, len(self.node_trajectories)):
            structure_df[f"x_{i}_mu"] = self.node_trajectories[i]["mu"]
            structure_df[f"x_{i}_pi"] = self.node_trajectories[i]["pi"]
            structure_df[f"x_{i}_muhat"] = self.node_trajectories[i]["muhat"]
            structure_df[f"x_{i}_pihat"] = self.node_trajectories[i]["pihat"]
            if (i == 1) & (self.model_type == "continuous"):
                surprise = gaussian_surprise(
                    x=self.node_trajectories[0]["value"][1:],
                    muhat=self.node_trajectories[i]["muhat"][:-1],
                    pihat=self.node_trajectories[i]["pihat"][:-1],
                )
            else:
                surprise = gaussian_surprise(
                    x=self.node_trajectories[i]["mu"][1:],
                    muhat=self.node_trajectories[i]["muhat"][:-1],
                    pihat=self.node_trajectories[i]["pihat"][:-1],
                )
            # fill with nans when the model cannot fit
            surprise = jnp.where(
                jnp.isnan(self.node_trajectories[i]["mu"][1:]), jnp.nan, surprise
            )
            structure_df[f"x_{i}_surprise"] = np.insert(surprise, 0, np.nan)

        # compute the global surprise over all node
        structure_df["surprise"] = structure_df.iloc[
            :, structure_df.columns.str.contains("_surprise")
        ].sum(axis=1, min_count=1)

        return structure_df

    def add_input_node(
        self,
        kind: str,
        input_idx: int = 0,
        omega_input: Union[float, np.ndarray, ArrayLike] = log(1e-4),
        pihat: Union[float, np.ndarray, ArrayLike] = jnp.inf,
        eta0: Union[float, np.ndarray, ArrayLike] = 0.0,
        eta1: Union[float, np.ndarray, ArrayLike] = 1.0,
    ):
        """Create an input node.

        Parameters
        ----------
        kind :
            The kind of input that should be created (can be `"continuous"` or
            `"binary"`).
        input_idx :
            The index of the new input (defaults to `0`).
        omega_input :
            The input precision (only relevant if `kind="continuous"`).
        pihat :
            The input precision (only relevant if `kind="binary"`).
        eta0 :
            The lower bound of the binary process (only relevant if `kind="binary"`).
        eta1 :
            The lower bound of the binary process (only relevant if `kind="binary"`).

        """
        if kind == "continuous":
            input_node_parameters = {
                "kappas": None,
                "omega": omega_input,
                "time_step": jnp.nan,
                "value": jnp.nan,
            }
        elif kind == "binary":
            input_node_parameters = {
                "pihat": pihat,
                "eta0": eta0,
                "eta1": eta1,
                "surprise": jnp.nan,
                "time_step": jnp.nan,
                "value": jnp.nan,
            }
        if input_idx == 0:
            # this is the first node, create the node structure
            self.parameters_structure = {input_idx: input_node_parameters}
            self.node_structure = (Indexes(None, None),)
            self.input_nodes_idx = (InputIndexes(input_idx, kind),)
        else:
            # update the node structure
            self.parameters_structure[input_idx] = input_node_parameters
            self.node_structure += (Indexes(None, None),)
            self.input_nodes_idx += (InputIndexes(input_idx, kind),)
        return self

    def add_value_parent(
        self,
        children_idxs: List,
        value_coupling: Union[float, np.ndarray, ArrayLike] = 1.0,
        mu: Union[float, np.ndarray, ArrayLike] = 0.0,
        mu_hat: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        pi: Union[float, np.ndarray, ArrayLike] = 1.0,
        pi_hat: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        kappas: Optional[Tuple] = None,
        nu: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        psis: Optional[Tuple] = None,
        omega: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        rho: Union[float, np.ndarray, ArrayLike] = 0.0,
    ):
        """Add a value parent to node.

        Parameters
        ----------
        children_idxs :
            The child(s) node index(es).
        value_coupling :
            The value_coupling between the child and parent node. This is will be
            appended to the `psis` parameters in the child node.
        mu :
            Mean.
        mu_hat :
            Expected mean.
        pi :
            Precision.
        pi_hat :
            Expected precision.
        kappas :
            Phasic part of the variance.
        nu :
            Stochasticity coupling.
        psis :
            Value coupling.
        omega :
            Tonic part of the variance.
        rho :
            Drift of the random walk.

        """
        # how many nodes in structure
        n_nodes = len(self.node_structure)
        parent_idx = n_nodes  # append a new parent in the structure

        # parent's parameter
        node_parameters = {
            "mu": mu,
            "muhat": mu_hat,
            "pi": pi,
            "pihat": pi_hat,
            "kappas": kappas,
            "nu": nu,
            "psis": psis,
            "omega": omega,
            "rho": rho,
        }

        # add a new node to connection structure with no parents
        self.node_structure += (Indexes(None, None),)

        # add a new parameters to parameters structure
        self.parameters_structure[parent_idx] = node_parameters

        # convert the structure to a list to modify it
        structure_as_list: List[Indexes] = list(self.node_structure)

        for idx in children_idxs:
            # add this node as parent and set value coupling
            if structure_as_list[idx].value_parents is not None:
                # append new index
                new_value_parents_idx = structure_as_list[idx].value_parents + (
                    parent_idx,
                )
                structure_as_list[idx] = Indexes(
                    new_value_parents_idx, structure_as_list[idx].volatility_parents
                )
                # set the value coupling strength
                self.parameters_structure[idx]["psis"] += (value_coupling,)
            else:
                # append new index
                new_value_parents_idx = (parent_idx,)
                structure_as_list[idx] = Indexes(
                    new_value_parents_idx, structure_as_list[idx].volatility_parents
                )
                # set the value coupling strength
                self.parameters_structure[idx]["psis"] = (value_coupling,)

        # convert the list back to a tuple
        self.node_structure = tuple(structure_as_list)

        return self

    def add_volatility_parent(
        self,
        children_idxs: List,
        volatility_coupling: Union[float, np.ndarray, ArrayLike] = 1.0,
        mu: Union[float, np.ndarray, ArrayLike] = 0.0,
        mu_hat: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        pi: Union[float, np.ndarray, ArrayLike] = 1.0,
        pi_hat: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        kappas: Optional[Tuple] = None,
        nu: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        psis: Optional[Tuple] = None,
        omega: Union[float, np.ndarray, ArrayLike] = jnp.nan,
        rho: Union[float, np.ndarray, ArrayLike] = 0.0,
    ):
        """Add a volatility parent to node.

        Parameters
        ----------
        children_idxs :
            The child(s) node index(es).
        volatility_coupling :
            The volatility coupling between the child and parent node. This is will be
            appended to the `kappas` parameters in the child node.
        mu :
            Mean.
        mu_hat :
            Expected mean.
        pi :
            Precision.
        pi_hat :
            Expected precision.
        kappas :
            Phasic part of the variance.
        nu :
            Stochasticity coupling.
        psis :
            Value coupling.
        omega :
            Tonic part of the variance.
        rho :
            Drift of the random walk.

        """
        # how many nodes in structure
        n_nodes = len(self.node_structure)
        parent_idx = n_nodes  # append a new parent in the structure

        # parent's parameter
        node_parameters = {
            "mu": mu,
            "muhat": mu_hat,
            "pi": pi,
            "pihat": pi_hat,
            "kappas": kappas,
            "nu": nu,
            "psis": psis,
            "omega": omega,
            "rho": rho,
        }

        # add a new node to connection structure with no parents
        self.node_structure += (Indexes(None, None),)

        # add a new parameters to parameters structure
        self.parameters_structure[parent_idx] = node_parameters

        # convert the structure to a list to modify it
        structure_as_list: List[Indexes] = list(self.node_structure)

        for idx in children_idxs:
            # add this node as parent and set value coupling
            if structure_as_list[idx].volatility_parents is not None:
                # append new index
                new_volatility_parents_idx = structure_as_list[
                    idx
                ].volatility_parents + (parent_idx,)
                structure_as_list[idx] = Indexes(
                    structure_as_list[idx].value_parents, new_volatility_parents_idx
                )
                # set the value coupling strength
                self.parameters_structure[idx]["kappas"] += (volatility_coupling,)
            else:
                # append new index
                new_volatility_parents_idx = (parent_idx,)
                structure_as_list[idx] = Indexes(
                    structure_as_list[idx].value_parents, new_volatility_parents_idx
                )
                # set the value coupling strength
                self.parameters_structure[idx]["kappas"] = (volatility_coupling,)

        # convert the list back to a tuple
        self.node_structure = tuple(structure_as_list)

        return self

    def get_update_sequence(
        self, node_idx: int = 0, update_sequence: Tuple = ()
    ) -> Tuple:
        """Get update sequence from the nodes' connections structure.

        Parameters
        ----------
        node_idx :
            The index of the current node.
        update_sequence :
            The current update sequence. This is used for recursive evaluation.

        Returns
        -------
        update_sequence :
            The update sequence.

        """
        # get the parents indexes
        value_parent_idx = self.node_structure[node_idx].value_parents
        volatility_parent_idx = self.node_structure[node_idx].volatility_parents

        # select the update function
        if node_idx == 0:
            if self.model_type == "binary":
                update_fn = binary_input_update
            elif self.model_type == "continuous":
                update_fn = continuous_input_update
        elif (node_idx == 1) & (self.model_type == "binary"):
            update_fn = binary_node_update
        else:
            update_fn = continuous_node_update

        # create a new sequence step and add it to the list
        new_sequence = node_idx, update_fn
        update_sequence += (new_sequence,)

        # search recursively for the nex update steps
        if value_parent_idx is not None:
            for idx in value_parent_idx:
                update_sequence = self.get_update_sequence(
                    node_idx=idx, update_sequence=update_sequence
                )
        if volatility_parent_idx is not None:
            for idx in volatility_parent_idx:
                update_sequence = self.get_update_sequence(
                    node_idx=idx, update_sequence=update_sequence
                )

        return update_sequence



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import Callable, NamedTuple, Optional, Tuple


class Indexes(NamedTuple):
    """Indexes to a node's value and volatility parents."""

    value_parents: Optional[Tuple]
    volatility_parents: Optional[Tuple]


class InputIndexes(NamedTuple):
    """Input node type and index."""

    idx: int
    kind: str


NodeStructure = Tuple[Indexes, ...]

UpdateSequence = Tuple[Tuple[int, Callable], ...]



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import unittest
from unittest import TestCase
import jax.numpy as jnp
from jax.lax import scan

from pyhgf import load_data
from pyhgf.continuous import (
    gaussian_surprise,
    continuous_input_update,
    continuous_node_update,
)
from pyhgf.structure import loop_inputs, apply_sequence
from jax.tree_util import Partial
from pyhgf.typing import Indexes

class Testcontinuous(TestCase):

    def test_update_sequences(self):

        # create a node structure with one value parent and one volatility parent
        node_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": None,
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 1.0,
        }
        input_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": (1.0,),
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 1.0,
        }
        node_structure = (
            Indexes(None, None),
            Indexes(None, None),
            Indexes(None, None),
        )

        parameters_structure = (
            input_parameters,
            node_parameters,
            node_parameters,
        )

        ###########################################
        # No value parent - no volatility parents #
        ###########################################
        sequence1 = 0, continuous_node_update
        update_sequence = (sequence1,)
        new_parameters_structure = apply_sequence(
            parameters_structure=parameters_structure, 
            time_step=1.0,
            node_structure=node_structure,
            update_sequence=update_sequence, 
            value=None
            )

        assert parameters_structure == new_parameters_structure

        #######################
        # x_1 as value parent #
        #######################
        node_structure = (
            Indexes((1,), None),
            Indexes(None, None),
            Indexes(None, None),
        )

        sequence1 = 0, continuous_node_update
        update_sequence = (sequence1,)
        new_parameters_structure = apply_sequence(
            parameters_structure=parameters_structure, 
            time_step=1.0,
            node_structure=node_structure,
            update_sequence=update_sequence, 
            value=None
            )

        assert jnp.isclose(
            new_parameters_structure[1]["pi"],
            1.2689414
        )

        ############################
        # x_1 as volatility parent #
        ############################
        node_structure = (
            Indexes(None, (1,)),
            Indexes(None, None),
            Indexes(None, None),
        )

        sequence1 = 0, continuous_node_update
        update_sequence = (sequence1,)
        new_node_structure = apply_sequence(
            parameters_structure=parameters_structure, 
            time_step=1.0,
            node_structure=node_structure,
            update_sequence=update_sequence, 
            value=None
            )
        assert jnp.isclose(
            new_node_structure[1]["pi"],
            0.7689414
        )

        #####################################
        # Both value and volatility parents #
        #####################################
        node_structure = (
            Indexes((1,), (2,)),
            Indexes(None, None),
            Indexes(None, None),
        )

        sequence1 = 0, continuous_node_update
        update_sequence = (sequence1,)
        new_node_structure = apply_sequence(
            parameters_structure=parameters_structure, 
            time_step=1.0,
            node_structure=node_structure,
            update_sequence=update_sequence, 
            value=None
            )
        assert jnp.isclose(
            new_node_structure[1]["pi"],
            1.2689414
        )
        assert jnp.isclose(
            new_node_structure[2]["pi"],
            0.7689414
        )

    def test_gaussian_surprise(self):
        surprise = gaussian_surprise(
            x=jnp.array([1.0, 1.0]),
            muhat=jnp.array([0.0, 0.0]),
            pihat=jnp.array([1.0, 1.0]),
        )
        assert jnp.all(jnp.isclose(surprise, 1.4189385))

    def test_update_continuous_input_parents(self):

        ###############################################
        # one value parent with one volatility parent #
        ###############################################
        input_node_parameters = {
            "omega": 1.0,
            "kappas": None,
            "psis": None,
            "time_step": 0.0,
            "value": 0.0,
        }
        node_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": (1.0,),
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 1.0,
        }

        node_structure = (
            Indexes((1,), None),
            Indexes(None, (2,)),
            Indexes(None, None),
        )
        parameters_structure = (
            input_node_parameters,
            node_parameters,
            node_parameters,
        )
        
        # create update sequence
        sequence1 = 0, continuous_input_update
        sequence2 = 1, continuous_node_update
        update_sequence = (sequence1, sequence2)

        # apply sequence
        new_parameters_structure = apply_sequence(
            node_structure=node_structure,
            parameters_structure=parameters_structure,
            update_sequence=update_sequence, 
            time_step=1.0,
            value=.2
            )

        assert new_parameters_structure[1]["pi"] == 0.48708236

    def test_scan_loop(self):

        timeserie = load_data("continuous")

        # Create the data (value and time steps vectors)
        data = jnp.array([timeserie, jnp.ones(len(timeserie), dtype=int)]).T

        ###############################################
        # one value parent with one volatility parent #
        ###############################################
        input_node_parameters = {
            "omega": 1.0,
            "kappas": None,
            "psis": None,
            "time_step": 0.0,
            "value": 0.0,
        }
        node_parameters = {
            "pihat": 1.0,
            "pi": 1.0,
            "muhat": 1.0,
            "kappas": (1.0,),
            "mu": 1.0,
            "nu": 1.0,
            "psis": (1.0,),
            "omega": 1.0,
            "rho": 1.0,
        }

        node_structure = (
            Indexes((1,), None),
            Indexes(None, (2,)),
            Indexes(None, None),
        )
        parameters_structure = (
            input_node_parameters,
            node_parameters,
            node_parameters,
        )
        
        # create update sequence
        sequence1 = 0, continuous_input_update
        sequence2 = 1, continuous_node_update
        update_sequence = (sequence1, sequence2)

        # create the function that will be scaned
        scan_fn = Partial(
            loop_inputs, 
            update_sequence=update_sequence, 
            node_structure=node_structure
            )

        # Run the entire for loop
        last, final = scan(scan_fn, parameters_structure, data)


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

import unittest
from unittest import TestCase

import jax.numpy as jnp

from pyhgf import load_data
from pyhgf.model import HGF


class Testmodel(TestCase):
    def test_HGF(self):
        """Test the model class"""

        ##############
        # Continuous #
        ##############
        timeserie = load_data("continuous")

        # two-level
        # ---------
        two_level_continuous_hgf = HGF(
            n_levels=2,
            model_type="continuous",
            initial_mu={"1": timeserie[0], "2": 0.0},
            initial_pi={"1": 1e4, "2": 1e1},
            omega={"1": -3.0, "2": -3.0},
            rho={"1": 0.0, "2": 0.0},
            kappas={"1": 1.0},
        )
        
        two_level_continuous_hgf.input_data(input_data=timeserie)

        surprise = two_level_continuous_hgf.surprise()  # Sum the surprise for this model
        assert jnp.isclose(surprise, -676.51306)
        assert len(two_level_continuous_hgf.node_trajectories[1]["mu"]) == 614

        # three-level
        # -----------
        three_level_continuous_hgf = HGF(
            n_levels=3,
            model_type="continuous",
            initial_mu={"1": 1.04, "2": 1.0, "3": 1.0},
            initial_pi={"1": 1e4, "2": 1e1, "3": 1e1},
            omega={"1": -13.0, "2": -2.0, "3": -2.0},
            rho={"1": 0.0, "2": 0.0, "3": 0.0},
            kappas={"1": 1.0, "2": 1.0},
        )
        three_level_continuous_hgf.input_data(input_data=timeserie)
        surprise = three_level_continuous_hgf.surprise()
        assert jnp.isclose(surprise, -394.20514)


        ##########
        # Binary #
        ##########
        timeseries = load_data("binary")

        # two-level
        # ---------
        two_level_binary_hgf = HGF(
            n_levels=2,
            model_type="binary",
            initial_mu={"1": .0, "2": .5},
            initial_pi={"1": .0, "2": 1e4},
            omega={"1": None, "2": -6.0},
            rho={"1": None, "2": 0.0},
            kappas={"1": None},
            eta0=0.0,
            eta1=1.0,
            pihat = jnp.inf,
        )

        # Provide new observations
        two_level_binary_hgf = two_level_binary_hgf.input_data(timeseries)
        surprise = two_level_binary_hgf.surprise()
        assert jnp.isclose(surprise, 215.58821)

        # three-level
        # -----------
        three_level_binary_hgf = HGF(
            n_levels=3,
            model_type="binary",
            initial_mu={"1": .0, "2": .5, "3": 0.},
            initial_pi={"1": .0, "2": 1e4, "3": 1e1},
            omega={"1": None, "2": -6.0, "3": -2.0},
            rho={"1": None, "2": 0.0, "3": 0.0},
            kappas={"1": None, "2": 1.0},
            eta0=0.0,
            eta1=1.0,
            pihat=jnp.inf,
        )
        three_level_binary_hgf.input_data(input_data=timeseries)
        surprise = three_level_binary_hgf.surprise()
        assert jnp.isclose(surprise, 215.59067)


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from typing import TYPE_CHECKING

import jax.numpy as jnp

from pyhgf.continuous import gaussian_surprise

if TYPE_CHECKING:
    from pyhgf.model import HGF


def first_level_gaussian_surprise(hgf: "HGF", response_function_parameters=None):
    """Gaussian surprise at the first level of a probabilistic network.

    .. note::
      The Gaussian surprise at the first level is the default method to compute surprise
      for continuous models. The function returns `jnp.inf` if the model could not fit
      at a given time point.

    Parameters
    ----------
    hgf :
        Instance of the HGF model.
    response_function_parameters :
        No additional parameters are required to compute the Gaussian surprise.

    Returns
    -------
    surprise :
        The model surprise given the input data.

    """
    # compute the sum of Gaussian surprise at the first level
    # the input value at time t is compared to the gaussian prediction at t-1
    surprise = jnp.sum(
        gaussian_surprise(
            x=hgf.node_trajectories[0]["value"][1:],
            muhat=hgf.node_trajectories[1]["muhat"][:-1],
            pihat=hgf.node_trajectories[1]["pihat"][:-1],
        )
    )

    # Return an infinite surprise if the model could not fit at any point
    return jnp.where(
        jnp.any(jnp.isnan(hgf.node_trajectories[1]["mu"])), jnp.inf, surprise
    )


def first_level_binary_surprise(hgf: "HGF", response_function_parameters=None):
    """Sum of the binary surprise along the time series (binary HGF).

    .. note::
      The binary surprise is the default method to compute surprise when
      `model_type=="binary"`, therefore this method will only return the sum of
      valid time points, and `jnp.inf` if the model could not fit.

    Parameters
    ----------
    hgf :
        Instance of the HGF model.
    response_function_parameters :
        No additional parameters are required to compute the binary surprise.

    Returns
    -------
    surprise :
        The model surprise given the input data.

    """
    # Return an infinite surprise if the model cannot fit
    this_surprise = jnp.where(
        jnp.isnan(hgf.node_trajectories[0]["surprise"]),
        jnp.inf,
        hgf.node_trajectories[0]["surprise"],
    )

    # Sum the surprise for this model
    surprise = jnp.sum(this_surprise)

    return surprise



# Author: Nicolas Legrand <nicolas.legrand@cas.au.dk>

from functools import partial
from typing import Dict, Union

import jax.numpy as jnp
from jax import Array, jit
from jax.lax import cond
from jax.typing import ArrayLike

from pyhgf.typing import NodeStructure


@partial(jit, static_argnames=("node_structure", "node_idx"))
def binary_node_update(
    parameters_structure: Dict,
    time_step: float,
    node_idx: int,
    node_structure: NodeStructure,
    **args
) -> Dict:
    """Update the value and volatility parents of a binary node.

    If the parents have value and/or volatility parents, they will be updated
    recursively.

    Updating the node's parents is a two step process:
        1. Update value parent(s) and their parents (if provided).
        2. Update volatility parent(s) and their parents (if provided).

    Then returns the new node tuple `(parameters, value_parents, volatility_parents)`.

    Parameters
    ----------
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    time_step :
        Interval between the previous time point and the current time point.
    node_idx :
        Pointer to the node that need to be updated. After continuous update, the
        parameters of value and volatility parents (if any) will be different.
    node_structure :
        Tuple of :py:class:`pyhgf.typing.Indexes` with same length than number of node.
        For each node, the index list value and volatility parents.

    Returns
    -------
    parameters_structure :
        The updated node structure.

    """
    # using the current node index, unwrap parameters and parents
    node_parameters = parameters_structure[node_idx]
    value_parents_idx = node_structure[node_idx].value_parents
    volatility_parents_idx = node_structure[node_idx].volatility_parents

    # Return here if no parents node are provided
    if (value_parents_idx is None) and (volatility_parents_idx is None):
        return parameters_structure

    pihat = node_parameters["pihat"]
    vape = jnp.subtract(node_parameters["mu"], node_parameters["muhat"])

    #######################################
    # Update the continuous value parents #
    #######################################
    if value_parents_idx is not None:
        # unpack the current parent's parameters with value and volatility parents
        va_pa_node_parameters = parameters_structure[value_parents_idx[0]]
        # va_pa_value_parents_idx = node_structure[value_parents_idx[0]].value_parents
        va_pa_volatility_parents_idx = node_structure[
            value_parents_idx[0]
        ].volatility_parents

        # 1. get pihat_pa and nu_pa from the value parent (x2)

        # 1.1 get new_nu (x2)
        # 1.1.1 get logvol
        logvol = va_pa_node_parameters["omega"]

        # 1.1.2 Look at the (optional) va_pa's volatility parents
        # and update logvol accordingly
        if va_pa_volatility_parents_idx is not None:
            for va_pa_vo_pa, k in zip(
                va_pa_volatility_parents_idx, va_pa_node_parameters["kappas"]
            ):
                logvol += k * parameters_structure[va_pa_vo_pa]["mu"]

        # 1.1.3 Compute new_nu
        nu = time_step * jnp.exp(logvol)
        new_nu = jnp.where(nu > 1e-128, nu, jnp.nan)

        # 1.2 Compute new value for nu and pihat
        pihat_pa, nu_pa = [1 / (1 / va_pa_node_parameters["pi"] + new_nu), new_nu]

        # 2.
        pi_pa = pihat_pa + 1 / pihat

        # 3. get muhat_pa from value parent (x2)

        # 3.1
        driftrate = va_pa_node_parameters["rho"]

        # TODO: this will be decided once we figure out the multi parents/children
        # 3.2 Look at the (optional) va_pa's value parents
        # and update driftrate accordingly
        # if va_pa_value_parents is not None:
        #     psi
        #     for va_pa_va_pa, psi in zip(
        #         va_pa_value_parents, va_pa_node_parameters["psis"]
        #     ):
        #         _mu = va_pa_va_pa[0]["mu"]
        #         driftrate += psi * _mu
        # 3.3
        muhat_pa = va_pa_node_parameters["mu"] + time_step * driftrate

        # 4.
        mu_pa = muhat_pa + vape / pi_pa  # This line differs from the continuous input

        # 5. Update node's parameters and node's parents recursively
        parameters_structure[value_parents_idx[0]]["pihat"] = pihat_pa
        parameters_structure[value_parents_idx[0]]["pi"] = pi_pa
        parameters_structure[value_parents_idx[0]]["muhat"] = muhat_pa
        parameters_structure[value_parents_idx[0]]["mu"] = mu_pa
        parameters_structure[value_parents_idx[0]]["nu"] = nu_pa

    return parameters_structure


@partial(jit, static_argnames=("node_structure", "node_idx"))
def binary_input_update(
    parameters_structure: Dict,
    time_step: float,
    node_idx: int,
    node_structure: NodeStructure,
    value: float,
) -> Dict:
    """Update the input node structure given one observation.

    This function is the entry level of the model fitting. It update the partents of
    the input node and then call :py:func:`pyhgf.binary.binary_node_update` to update
    the rest of the node structure.

    Parameters
    ----------
    value :
        The new observed value.
    time_step :
        Interval between the previous time point and the current time point.
    parameters_structure :
        The structure of nodes' parameters. Each parameter is a dictionary with the
        following parameters: `"pihat", "pi", "muhat", "mu", "nu", "psis", "omega"` for
        continuous nodes.
    .. note::
        `"psis"` is the value coupling strength. It should have same length than the
        volatility parents' indexes. `"kappas"` is the volatility coupling strength.
        It should have same length than the volatility parents' indexes.
    node_structure :
        Tuple of :py:class:`pyhgf.typing.Indexes` with same length than number of node.
        For each node, the index list value and volatility parents.
    node_idx :
        Pointer to the node that need to be updated. After continuous update, the
        parameters of value and volatility parents (if any) will be different.

    Returns
    -------
    parameters_structure :
        The updated parameters structure.

    See Also
    --------
    update_continuous_parents, update_continuous_input_parents

    """
    # using the current node index, unwrap parameters and parents
    input_node_parameters = parameters_structure[node_idx]
    value_parents_idx = node_structure[node_idx].value_parents
    volatility_parents_idx = node_structure[node_idx].volatility_parents

    if (value_parents_idx is None) and (volatility_parents_idx is None):
        return parameters_structure

    pihat = input_node_parameters["pihat"]

    ################################
    # Update parents (binary node) #
    ################################

    if value_parents_idx is not None:
        # unpack the current parent's parameters with value and volatility parents
        # va_pa_node_parameters = parameters_structure[value_parents_idx[0]]
        va_pa_value_parents_idx = node_structure[value_parents_idx[0]].value_parents
        # va_pa_volatility_parents_idx = node_structure[
        #     value_parents_idx[0]
        # ].volatility_parents

        # 1. Compute new muhat_pa and pihat_pa from binary node parent
        # ------------------------------------------------------------

        # 1.1 Compute new_muhat from continuous node parent (x2)

        # 1.1.1 get rho from the value parent of the binary node (x2)
        driftrate = parameters_structure[va_pa_value_parents_idx[0]]["rho"]

        # TODO: this will be decided once we figure out the multi parents/children
        # # 1.1.2 Look at the (optional) va_pa's value parents (x3)
        # # and update the driftrate accordingly
        # if va_pa_value_parents[0][1] is not None:
        #     for va_pa_va_pa in va_pa_value_parents[0][1]:
        #         # For each x2's value parents (optional)
        #         _psi = va_pa_value_parents[0][0]["psis"]
        #         _mu = va_pa_va_pa[0]["mu"]
        #         driftrate += _psi * _mu

        # 1.1.3 compute new_muhat
        muhat_va_pa = (
            parameters_structure[va_pa_value_parents_idx[0]]["mu"]
            + time_step * driftrate
        )

        muhat_va_pa = sgm(muhat_va_pa)
        pihat_va_pa = 1 / (muhat_va_pa * (1 - muhat_va_pa))

        # 2. Compute surprise
        # -------------------
        eta0 = input_node_parameters["eta0"]
        eta1 = input_node_parameters["eta1"]

        mu_va_pa, pi_va_pa, surprise = cond(
            pihat == jnp.inf,
            input_surprise_inf,
            input_surprise_reg,
            (pihat, value, eta1, eta0, muhat_va_pa),
        )

        # Update value parent's parameters
        parameters_structure[value_parents_idx[0]]["pihat"] = pihat_va_pa
        parameters_structure[value_parents_idx[0]]["pi"] = pi_va_pa
        parameters_structure[value_parents_idx[0]]["muhat"] = muhat_va_pa
        parameters_structure[value_parents_idx[0]]["mu"] = mu_va_pa

    parameters_structure[node_idx]["surprise"] = surprise
    parameters_structure[node_idx]["time_step"] = time_step
    parameters_structure[node_idx]["value"] = value

    return parameters_structure


def gaussian_density(x: ArrayLike, mu: ArrayLike, pi: ArrayLike):
    """Gaussian density as defined by mean and precision."""
    return (
        pi
        / jnp.sqrt(2 * jnp.pi)
        * jnp.exp(jnp.subtract(0, pi) / 2 * (jnp.subtract(x, mu)) ** 2)
    )


def sgm(
    x,
    lower_bound: Union[ArrayLike, float] = 0.0,
    upper_bound: Union[ArrayLike, float] = 1.0,
) -> Array:
    """Logistic sigmoid function."""
    return jnp.subtract(upper_bound, lower_bound) / (1 + jnp.exp(-x)) + lower_bound


def binary_surprise(
    x: Union[float, ArrayLike], muhat: Union[float, ArrayLike]
) -> Array:
    r"""Surprise at a binary outcome.

    The surprise ellicited by a binary observation :math:`x` mean :math:`\hat{\mu}`
    and expected probability :math:`\hat{\pi}` is given by:

    .. math::

       \begin{cases}
            -\log(\hat{\mu}),& \text{if } x=1\\
            -\log(1 - \hat{\mu}), & \text{if } x=0\\
        \end{cases}

    Parameters
    ----------
    x :
        The outcome.
    muhat :
        The mean of the Bernoulli distribution.

    Returns
    -------
    surprise :
        The binary surprise.


    Examples
    --------
    >>> from pyhgf.binary import binary_surprise
    >>> binary_surprise(x=1.0, muhat=0.7)
    `Array(0.35667497, dtype=float32, weak_type=True)`

    """
    return jnp.where(x, -jnp.log(muhat), -jnp.log(jnp.array(1.0) - muhat))


def input_surprise_inf(op):
    """Apply special case if pihat is `jnp.inf` (just pass the value through)."""
    (pihat, value, eta1, eta0, muhat_va_pa) = op
    mu_va_pa = value
    pi_va_pa = jnp.inf
    surprise = binary_surprise(value, muhat_va_pa)

    return mu_va_pa, pi_va_pa, surprise


def input_surprise_reg(op):
    """Compute the surprise, mu_va_pa and pi_va_pa if pihat is not `jnp.inf`."""
    (pihat, value, eta1, eta0, muhat_va_pa) = op

    # Likelihood under eta1
    und1 = jnp.exp(jnp.subtract(0, pihat) / 2 * (jnp.subtract(value, eta1)) ** 2)

    # Likelihood under eta0
    und0 = jnp.exp(jnp.subtract(0, pihat) / 2 * (jnp.subtract(value, eta0)) ** 2)

    # Eq. 39 in Mathys et al. (2014) (i.e., Bayes)
    mu_va_pa = muhat_va_pa * und1 / (muhat_va_pa * und1 + (1 - muhat_va_pa) * und0)
    pi_va_pa = 1 / (mu_va_pa * (1 - mu_va_pa))

    # Surprise
    surprise = -jnp.log(
        muhat_va_pa * gaussian_density(value, eta1, pihat)
        + (1 - muhat_va_pa) * gaussian_density(value, eta0, pihat)
    )

    return mu_va_pa, pi_va_pa, surprise
